import WebSocket, { WebSocketServer } from 'ws';
import { exec } from 'child_process';

const wss = new WebSocketServer({ port: 8080 });
const workers = [];
const queue = [];

const alphabet = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';

setInterval(() => {
    const freeWorkers = workers.filter((worker) => !worker.hash);
    freeWorkers.forEach((worker) => {
        worker.send('exit');
        worker.close();
    });
    workers.splice(0, freeWorkers.length);
}, 60_000);

wss.on('connection', (ws) => {
    ws.on('message', (message) => {
        console.log(`Received message => ${message}`);
        if(ws.firstMessage && message === 'slave') {
            workers.push(ws);
            ws.on('message', handleWorkerMessage);
            if(queue.length) {
                const { hash, needed, count, minLen, maxLen } = queue[0];
                const start = alphabet.length ** minLen;
                const end = alphabet.length ** maxLen;
                const interval = (end - start) / count;
                const currentIndex = count - needed + i;
                const currentStart = start + (interval * currentIndex);
                const currentEnd = start + (interval * (currentIndex + 1));
                ws.hash = hash;
                ws.send(`search ${hash} ${base10ToBase62(currentStart)} ${base10ToBase62(currentEnd)}`);
                queue[0] = {...queue[0], needed: needed - 1};
            }
        }else {
            ws.on('message', handleClientMessage);
        }
        ws.removeListener('message', this);
    });
});

function handleWorkerMessage(message) {
    if(!message.startsWith('found')) return;
    const [_, hash, password] = message.split(' ');
    thisHashWorkers = workers.filter((worker) => worker.hash === hash);
    thisHashWorkers.forEach((worker) => {
        worker.send(`stop`)
        worker.hash = null;
    });
}

function handleClientMessage(message) {
    const [hash, minLen, maxLen, workersCount] = message.split(':');
    const freeWorkers = workers.filter((worker) => !worker.hash);
    freeWorkers
    const missingWorkers = freeWorkers - workersCount;
    if(missingWorkers > 0) exec(`docker-compose up -d --scale worker=${workers.length + missingWorkers}`);
    queue.push({ hash, needed: missingWorkers, count: workersCount, minLen, maxLen });
}

function base10ToBase62(number) {
    let result = '';
    while(number > 0) {
        result = alphabet[number % alphabet.length] + result;
        number = Math.floor(number / alphabet.length);
    }
    return result;
} 